/*
 * readarkeia.c  ---  v8  ---  For ARKEIA  ----
 * Copyright (c) 1999-2009 ARKEIA
 * All rights reserved
 * Your licence agreement specifies the terms and conditions
 * for the use of this software
 */

/* NL_NOUPDATE */

const char readarkeia_version[] = "8.0.11";

/*
 * IMPORTANT NOTICE: The following program is provided by ARKEIA for emergency
 * case only. ARKEIA denies all warranties about the safety of this program for
 * data existing on the machine. No implicit or explicit promise is made by
 * ARKEIA about the quality, the result or conformity with user needs.
 * The use of this source program is made under the user supervision and under
 * his sole responsibility.
 *
 * ARKEIA DENIES ALL RESPONSIBILITY FOR USE OF THIS PROGRAM AND/OR ITS RESULTS.
 */

/*
 * READARKEIA: This program allows the restoration of backup tapes made with
 *          ARKEIA.
 *
 *
 * You can use readarkeia like it was a tar command. It's the same syntax.
 * Not all options are implemented.
 *
 * SYNOPSIS
 *   readarkeia [-l<loglevel>] x|t|i [mfuv] [device|-] [filename1 filename2 ...]
 *
 * DESCRIPTION
 *     -l   Change the verbosity level. Use it to obtain detailed
 *          information on what is really on the tape. Value can be from
 *          10 to 80.
 *
 *     t    List the table of contents of the tape.
 *
 *     x    Extract files from the tape. If no files are specified, all
 *          the file on the tape are extracted.
 *
 *     i    Get information about the tape.
 *
 *     m    This option tells readarkeia that we are on the middle of
 *          the tape. (There is no label to read).
 *
 *     f    Use the next argument as the name of the device. If device name
 *          is given as `-', readarkeia reads the tape date from the
 *          standard input.
 *
 *     b    Use the next argument as the block size in bytes. 
 *          This is usefull for LTO3 and LTO4 tape drives. Values are
 *          65536 for LTO3 and 131072 for LTO4
 *
 *     u    This option updates the file being restored to the newest
 *          version found on a tape. Without using this option, if
 *          several versions of the same file are found on the same
 *          tape, all versions are restored with a different numerical
 *          extension. For example foo.1, foo.2, foo.3, and so on. When
 *          the -u option is specified, only the newest version of foo
 *          is restored.
 *
 *     v    Verbose.
 *
 *     The filenames are optionnal. As said before, if no file is specified
 *     then the entire tape is restored. The filenames have the following
 *     syntax:
 *                <HOST>/<FILE_PATH or PATH_PART>
 *
 *     If you want to restore a complete path or several paths you can specify
 *     the common prefix. Example:
 *                The filename 'alpha/u' will match paths like
 *                    'alpha/usr', 'alpha/u3', 'alpha/users', ...
 *
 * EXAMPLES
 *     To get a table of contents of a tape put in the device '/dev/rmt0'
 *     on an AIX system. The command is:
 *              readarkeia tvf /dev/rmt0
 *
 *     Then if you want to extract file like '/usr/lib/libc.a' you will
 *     indicate at the beginning of the filename the hostname. So the filename
 *     will be: '<HOST>/usr/lib/libc.a'
 *     The command id:
 *              readarkeia xvf /dev/rmt0 alpha/usr/lib/libc.a
 *
 * WARNING
 *     readarkeia program will restore the newest version of a file, renaming
 *     silently any existing version.
 */

#ifdef NLO_WIN32
#	include <windows.h>
#	include <io.h>
#	include <sys/utime.h>
#	include <process.h>
#else
#	include <unistd.h>
#	include <utime.h>
#endif /* NLO_WIN32 */
#include <stdio.h>
#include <errno.h>
#include <time.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <malloc.h>
#include <stdlib.h>
#include <string.h>


#ifdef NLO_WIN32
#	define mode_t	int

#	define	S_ISDIR(mode)	(((mode) & S_IFMT) == S_IFDIR)
#	define	S_ISCHR(mode)	(((mode) & S_IFMT) == S_IFCHR)
#	define	S_ISBLK(mode)	(((mode) & S_IFMT) == S_IFBLK)
#	define	S_ISREG(mode)	((((mode) & S_IFMT) == S_IFREG)		\
				|| !((mode) & S_IFMT))
#	ifdef S_IFIFO
#		define	S_ISFIFO(mode)	(((mode) & S_IFMT) == S_IFIFO)
#	else
#		define S_ISFIFO(mode) (0)
#	endif
#	define S_ISUID		0004000
#	define S_ISGID		0002000
#	define S_ISVTX		0001000

#	define getpid	_getpid
#	define sleep(x)	Sleep(1000*(x))
#endif /* NLO_WIN32 */

#define INT4	int
#define UBYTE	unsigned char	/* Unsigned byte */
#define UCHAR	unsigned char
#define UINT4	unsigned int
#define UINT	unsigned int
#define ULONG	unsigned long	/* Unsigned word (4 bytes) */
#define UWORD	unsigned short	/* Unsigned word (2 bytes) */
#define UCARD	unsigned	/* Which ever is fastest, uns. or int  */


#define TEMP_DIR        	"."
#define STR_SIZE		1024
#define OUI			(-1)
#define NON			0

#ifndef MAX
# define MAX(a,b) ((a > b) ? a : b)
#endif

static int nl_loglevel = 0;	/* Default */
static int nl_line;

#define GOTO_ERROR		{nl_line=__LINE__; goto error;}
#define NL_LOG(level)		if(nl_loglevel>=(level)) printf
#define INFO(dummylevel)	if(global_operation==ADMIN_INFO) printf

#ifdef NLO_WIN32
#define NL_ERR(dummylevel)	printf("ERROR: "); printf
#else
#define NL_ERR(dummylevel)	\
	{ \
		 int local_errno; \
		 local_errno = errno; \
		 printf("ERROR: "); \
		 if (local_errno) { \
				 printf("(%d: %s) ", local_errno, \
						 strerror(local_errno)); \
		 } \
	 } \
	 printf
#endif /* NLO_WIN32 */

#ifdef NLO_WIN32
static LPVOID s_lpMsgBuf = 0;
#	define FillWin32Error()						\
		FormatMessage(FORMAT_MESSAGE_ALLOCATE_BUFFER		\
				| FORMAT_MESSAGE_FROM_SYSTEM,		\
			NULL, GetLastError(),				\
			MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT),	\
			(LPTSTR) &s_lpMsgBuf, 0, NULL );
#	define FreeWin32Error()		LocalFree(s_lpMsgBuf)
#	define GetWin32Error()		s_lpMsgBuf

#   define ark_read(fd, buf, len)            \
        ((pdev->dvtype == DV_FILE)      \
            ? wnt_read_file(fd, buf, len)   \
            : wnt_read(fd, buf, len))
#	define ark_close		CloseHandle
#	define ark_write(fd, pbuf, n) {					\
		DWORD written;						\
									\
		if (WriteFile(fd, pbuf, n, &written, NULL) == FALSE) {	\
			FillWin32Error();				\
			NL_ERR(0) ("Write error %d (%s).\n",		\
				GetLastError(), GetWin32Error());	\
			FreeWin32Error();				\
		}							\
	}
DWORD ark_lseek(HANDLE fd, off_t offset, int whence)
{
	DWORD method;
	DWORD low_offset;

	switch (whence) {
		case SEEK_SET:
			method = FILE_BEGIN;
			break;
		case SEEK_CUR:
			method = FILE_CURRENT;
			break;
		case SEEK_END:
			method = FILE_END;
			break;
		default:
			return -1;
			break;
	}

	/* this is limited to files smaller than 2GB */
	if ((low_offset = SetFilePointer(fd, (DWORD) offset, NULL, method))
			== INVALID_SET_FILE_POINTER) {
		FillWin32Error();
		NL_ERR(0) ("SetFilePointer error %d (%s)\n",
			GetLastError(), GetWin32Error());
		FreeWin32Error();
		return -1;
	}
	return low_offset;
}

HANDLE tape_open(char *device_name, int flags)
{
	HANDLE fd = INVALID_HANDLE_VALUE;
	char wnt_dev_name[256];

	sprintf(wnt_dev_name, "\\\\.\\%s", device_name);
	fd = CreateFile(
		wnt_dev_name,
		GENERIC_READ | GENERIC_WRITE,
		0,
		NULL,
		OPEN_EXISTING,
		FILE_FLAG_NO_BUFFERING | FILE_FLAG_WRITE_THROUGH,
		NULL);
	if (fd == INVALID_HANDLE_VALUE) {
		FillWin32Error();
		NL_ERR(0) ("Cannot open device '%s' (%d: '%s')\n",
			wnt_dev_name, GetLastError(), GetWin32Error());
		FreeWin32Error();
	}
	return fd;
}

int wnt_read_file(HANDLE fd, char *buf, int len)
{
	DWORD dwSize;


	if (!ReadFile(fd, buf, (DWORD) len, &dwSize, NULL )) {
		dwSize = -1;
		FillWin32Error();
		NL_ERR(0) ("ReadFile error: %d (%s)\n",
			GetLastError(), GetWin32Error());
		FreeWin32Error();
	}

	return (int) dwSize;
}

int wnt_read(HANDLE fd, char *buf, int len)
{
	TAPE_SET_MEDIA_PARAMETERS tpp;
	DWORD dwSize;
	int error;

	ZeroMemory(&tpp, sizeof(TAPE_SET_MEDIA_PARAMETERS));
	tpp.BlockSize = len;

	if (!(SetTapeParameters(
			fd, SET_TAPE_MEDIA_INFORMATION, &tpp)) == NO_ERROR) {
		error = GetLastError();
		if (error != 1) {
			FillWin32Error();
			NL_ERR(0) ("Error %d setting tape parameters (%s)\n",
				error, GetWin32Error());
			FreeWin32Error();
			return -1;
		}
	}

	if (!ReadFile(fd, buf, (DWORD) len, &dwSize, NULL )) {
		error = GetLastError();
		if (error == ERROR_FILEMARK_DETECTED) dwSize = 0;
		else dwSize = -1;
		FillWin32Error();
		NL_ERR(0) ("ReadFile error: %d (%s)\n", error, GetWin32Error());
		FreeWin32Error();
	}

	return (int) dwSize;
}

#else
#	define ark_read			read
#	define ark_close		close
#	define ark_write		write
#	define ark_lseek		lseek
#	define tape_open		open
#endif /* NLO_WIN32 */



extern int errno;

char *compress_type();
int hextoi();
void end_of_file();
void myexit();
void uncompress();
void usage();


#define F_ST			252		/* strategy process */
#define HEADER_SIZE		12
#define MAXFLOWS		200		/* max number of flows */
#define OPB_SERV_VERSION	"@(#)OPB-SVERSION"
#define OPB_VERSION		"@(#)OPB-VERSION!!"
#define TAPEPAGESIZE		(8*1024)

#ifdef NLO_WIN32
#	define INVALID_FD	INVALID_HANDLE_VALUE
#	define ARK_HANDLE	HANDLE
#else
#	define INVALID_FD	-1
#	define ARK_HANDLE	int
#endif /* NLO_WIN32 */

/*
 * Structure of a tape block
 * a seq number is 4 bits long, a number of 8k pages is 3 bits
 */

#define BLKSZMASK		7	/* 00000111 */
#define DB_BLOCK		8	/* 00001000 */
#define O3MAGICTAPE	0xec557948

struct stapblk {
	UINT4 blkseqnum;	/* flow[8] + seqnum */
	UINT4 blkbkid;		/* backup id */
	UCHAR blkdummy;		/* 8 spare bits */
	UCHAR blksz;	/* (spare bits <<4) | DB_BLOCK | Number of 8k pages */
	UCHAR blkproffh; /* M_BKDAT: high byte offset of first prop on block */
			/* M_DBDAT: 0xff if 1st block of db for slot, else 0 */
	UCHAR blkproffl; /* M_BKDAT: low byte offset of first prop on block */
			/* M_DBDAT: 0xff if last block of db for slot, else 0*/
	UCHAR blkdata[1];	/* The block data */
};

/* Macro to extract flow number from a composite number*/
#define COMP2FLOW(num) ((num>>24)&255)

struct stapblk *in_blk = 0;

struct st_flow {
	int	 fnum;
	UINT4	 bkid;
	UINT4	 blknum;
	char	*prop;
	char	*value;
	char	*realsize;
	int	 state;	/* For prop reading */
	int	 operation;
	int	 bin_rest;

	/* Prop storage */
	char	*fullname;
	char	*host;
	char	*volume;
	char	*plugin;
	char	*location;
	char	*component;
	char	*link_path;
	char	*obj_type;
	char	*slave_version;
	mode_t	 type;
	dev_t	 ark_dev;	/* device file resides on */
	ino_t	 ark_ino;	/* the file serial number */
	mode_t	 ark_mode;	/* file mode */
#ifndef NLO_WIN32
	nlink_t	 ark_nlink;	/* number of hard links to the file */
	uid_t	 ark_uid;	/* user ID of owner */
	gid_t	 ark_gid;	/* group ID of owner */
#endif /* !NLO_WIN32 */
	dev_t	 ark_rdev;	/* the device identifier (special files only)*/
	off_t	 ark_size;	/* total size of file, in bytes */
	time_t	 ark_atime;	/* file last access time */
	time_t	 ark_mtime;	/* file last modify time */
	time_t	 ark_ctime;	/* file last status change time */

	/* File section */
	char	*buff;
	int	 buff_size;
	int	 flags;			/* Bit mask */
	char	*cur_pc;
	ARK_HANDLE	  fd;
};

struct st_device {
	ARK_HANDLE  dvfd;      /* file descriptor */
	char    *dvname;       /* name of rewindable device */
	int     dvatbegseg;    /* 0 if read or write != 0 */
	int     dvlastread;    /* return of last read */
	int     dvcurseg;      /* current segment (>= 0) */
	int     dvtype;        /* DV_NORMAL | DV_FILE */
	char    *dvbuf;        /* buffer for read */
	int     dvbufsize;     /* size of buffer */
	int     dvinbuf;       /* size of buffer really read */
	int     dvbufread;     /* size of buffer read and returned */
};


/* Defines for <dvtype>
   -------------------- */
#define DV_NORMAL	0
#define DV_FILE		1

/* Defines for <state>
   -------------------- */
#define NORMAL_STATE		1
#define INCOMPLETE_PROP		2
#define INCOMPLETE_VAL		3
#define INCOMPLETE_REALSZ	4
#define INCOMPLETE_BIN		5

/* Defines for <operation>
   ----------------------- */
#define NO_OPERATION		1
#define EXTRACT_FILE		2
#define DONT_EXTRACT		3
#define CONSULT_LIST		4
#define ADMIN_INFO		5

/* Defines for <type>
   ------------------ */
#define OPB_BLKDEV		1
#define OPB_CHRDEV		2
#define OPB_DIR			3
#define OPB_FIFO		4
#define OPB_NAMEDFILE		5
#define OPB_OBJECT		6
#define OPB_RAWDEV		7
#define OPB_REGFILE		8
#define OPB_SYMLINK		9
#define OPB_UNKNOWN		10

/* Compression values */
#define COMPRESS_MASK		3
#define NO_COMPRESS		0
#define LZ1_COMPRESS		1
#define LZ3_3_COMPRESS		2

/* Defines for <flags>
   ------------------- */
#define COMPRESS_BUFF	(1<<1)
#define DONT_FREE_BUFF	(1<<2)
#define HARD_LINK	(1<<3)

/* Defines for <fd>
   ---------------- */
static ARK_HANDLE fd_in = INVALID_FD;

static struct st_device device_definition;
struct st_device *pdev = 0;

struct st_flow flow_tab[MAXFLOWS];

static int global_operation = NO_OPERATION;
static char *extract_path[1024];
static int nb_path = 0;

static short gl_update_mode = NON;  /* set to OUI if the 'u' option is used */


#define GETWORD(spc, val, mpc) \
	pc = getword(spc, val, mpc); \
	if (!pc) goto unexp_eob; \
	pc++

#define GETNCHARS(spc, val, mpc, n) \
	pc = getnchars(spc, val, mpc, n); \
	if (!pc) goto unexp_eob;

#define ALLOC_IF_NOT(ptr) \
	if (!(ptr)) { \
		if (!((ptr) = malloc(STR_SIZE+1))) { \
			NL_ERR(0)("Can't alloc memory for prop storage\n"); \
			GOTO_ERROR; \
		} \
		*(ptr) = 0; \
	 }


static int need_uint4_chsex = NON;

UINT4 ch_uint4_sex(in)
UINT4 in;
{
	UINT4 out;
	char *p1, *p2;

	need_uint4_chsex = OUI;
	p1 = (char *)&in;
	p2 = (char *)&out;
	p2[0] = p1[3];
	p2[1] = p1[2];
	p2[2] = p1[1];
	p2[3] = p1[0];
	return (out);
}


int match(filename)
char *filename;
{
	int i;
	char *pc1, *pc2;

	for (i = 0; i < nb_path; i++) {
		for (pc1 = filename, pc2 = extract_path[i];
		     *pc1 && *pc2 && (*pc1 == *pc2); pc1++, pc2++);
		if ((!*pc1 && !*pc2) || (*pc1 && !*pc2)) return (OUI);
	}
	return (NON);
}


void check_operation(pf)
struct st_flow *pf;
{
	char *pc1, *pc2;
	struct stat buf;
	int i;
	char str[1024];

	/* Make the full file name */
	pc2 = pf->fullname;
	for (pc1 = pf->host; *pc1; *pc2 = *pc1, pc1++, pc2++);
	if (*pf->plugin) {
		if (*(pc2 - 1) != '/') *pc2++ = '/';
		for (pc1 = pf->plugin; *pc1; *pc2 = *pc1, pc1++, pc2++);
	}
	if (*pf->volume) {
		if (*(pc2 - 1) != '/') *pc2++ = '/';
		for (pc1 = pf->volume; *pc1; *pc2 = *pc1, pc1++, pc2++);
	}
	if (*pf->location != '/') *pc2++ = '/';
	if ((strlen(pf->location) == 1) &&
	   (pf->location[0] == '/')) goto skip_location;
	for (pc1 = pf->location; *pc1; *pc2 = *pc1, pc1++, pc2++);
skip_location:
	if (*(pc2 - 1) != '/') *pc2++ = '/';
	for (pc1 = pf->component; *pc1; *pc2 = *pc1, pc1++, pc2++);
	if (*(pc2 - 1) == '/') pc2--;
	*pc2 = '\0';

	if (!*pf->location || !*pf->host || !*pf->fullname) {
		pf->operation = DONT_EXTRACT;
		goto its_ok;
	}

	pf->operation = global_operation;
	if ((global_operation != EXTRACT_FILE) || !nb_path) goto its_ok;

	/* EXTRACT_FILE */
	if (!match(pf->fullname)) pf->operation = DONT_EXTRACT;

its_ok:
	if (pf->operation == EXTRACT_FILE) {
		/* Check if this file already exist */
		if (stat(pf->fullname, &buf) == 0 && !S_ISDIR(buf.st_mode)) {
			if (gl_update_mode == NON) {
				i = 0;
				while (++i) {
					sprintf(str, "%s.%d", pf->fullname, i);
					if (stat(str, &buf)) {
						rename(pf->fullname, str);
						break;
					}
				}
			} else if (buf.st_mtime >= pf->ark_mtime) {
				pf->operation = DONT_EXTRACT;
			}
		}
	}
	return ;

}

/*
 * set access and modification times
 */
static void set_time(namep, atime, mtime)
register char *namep;
time_t atime, mtime;
{
	struct utimbuf times;

	times.actime = atime;
	times.modtime = mtime;
	utime(namep, &times);
}

/*
 * create recursively all dirs up to and including name
 */

int gendirs(name,including)
char *name;
int including; /* 0: do not create name, 1: create dir name */
{
	char *pc,*pc1;
	char str[1024];
	struct stat sta;
	static int recurs = -1;
	int domk;

	if (!*name) GOTO_ERROR;

	recurs++;
	if (!including) {
		strcpy(str, name);
		for (pc = pc1 = str; *pc; pc++) if (*pc == '/') pc1 = pc;
		*pc1 = 0;
		if (!gendirs(str,1)) GOTO_ERROR;
		goto retoui;
	}
	domk = 0;
	if (stat(name,&sta) < 0) domk = 1;
	else if (S_ISREG(sta.st_mode)) {
#ifdef NLO_WIN32
		if (DeleteFile(name) == FALSE) {
#else
		if (unlink(name) < 0) {
#endif /* NLO_WIN32 */
			NL_ERR(0)("Can't remove '%s'\n",name);
			GOTO_ERROR;
		}
		domk = 1;
	}
	if (domk) {
#ifdef NLO_WIN32
		if (CreateDirectory(name, 0) == FALSE) {
#else
		if (mkdir(name, 0755) < 0) {
#endif /* NLO_WIN32 */
			if (!gendirs(name, 0)) GOTO_ERROR;
#ifdef NLO_WIN32
			if (CreateDirectory(name, 0) == FALSE) {
#else
			if (mkdir(name, 0755) < 0) {
#endif /* NLO_WIN32 */
				NL_ERR(0)("Can't create directory '%s'\n", name);
				GOTO_ERROR;
			}
		}
	}
retoui:
	recurs--;
	return (OUI);
error:
	NL_ERR(0)("Error for directory '%s'\n", name);
	recurs--;
	return (NON);
}


int create_file(pf)
struct st_flow *pf;
{
#ifdef NLO_WIN32
	pf->fd = INVALID_FD;
	if (!*pf->fullname) GOTO_ERROR;

	if (!gendirs(pf->fullname, NON)) {
		NL_ERR(0)("Can't create path for file %s\n", pf->fullname);
		GOTO_ERROR;
	}
	switch (pf->type) {
	case OPB_DIR:
		if (!(pf->ark_mode & ~S_IFMT)) pf->ark_mode = 0755;
		if (CreateDirectory(pf->fullname, 0) == FALSE) {
			FillWin32Error();
			NL_ERR(0) ("Can't create directory %s (%d: '%s')\n",
				pf->fullname, GetLastError(),
				GetWin32Error());
				FreeWin32Error();
			GOTO_ERROR;
		}
		break;
	case OPB_NAMEDFILE:	/* Try to restore if there is data ??? */
	case OPB_OBJECT:	/* We create a file with object data */
	case OPB_RAWDEV:	/* We create a file with raw data */
	case OPB_REGFILE:
		if (pf->flags & HARD_LINK) {		/* Case of hard link */
			NL_ERR(0)("Can't link %s to %s (not supported on this platform)\n", pf->fullname, pf->link_path);
			GOTO_ERROR;
		}
		if (!(pf->ark_mode & ~S_IFMT)) pf->ark_mode = 0644;
		if ((pf->fd = CreateFile(pf->fullname,
				GENERIC_WRITE, 0, NULL,
				CREATE_ALWAYS, FILE_FLAG_BACKUP_SEMANTICS
					| FILE_FLAG_SEQUENTIAL_SCAN,
				NULL)) == INVALID_HANDLE_VALUE) {
			FillWin32Error();
			NL_ERR(0) ("Can't create file %s (%d: '%s')\n",
				pf->fullname, GetLastError(),
				GetWin32Error());
			FreeWin32Error();
			GOTO_ERROR;
		}
		break;
	default:
		NL_ERR(0)("Unknown file type for file %s\n", pf->fullname);
		GOTO_ERROR;
	}
	return OUI;
error:
	pf->operation = DONT_EXTRACT;
	return NON;
#else
	mode_t mode;

	pf->fd = INVALID_FD;
	if (!*pf->fullname) GOTO_ERROR;

	if (!gendirs(pf->fullname, NON)) {
		NL_ERR(0)("Can't create path for file %s\n", pf->fullname);
		GOTO_ERROR;
	}
	switch (pf->type) {
	case OPB_BLKDEV: mode = S_IFBLK; goto create_special;
	case OPB_CHRDEV: mode = S_IFCHR; goto create_special;
	case OPB_FIFO:	 mode = S_IFIFO; goto create_special;
	case OPB_DIR:
		if (!(pf->ark_mode & ~S_IFMT)) pf->ark_mode = 0755;
		if (mkdir(pf->fullname, pf->ark_mode) < 0) {
			if (errno == EEXIST) goto dir_ok; /* File exists */
			NL_ERR(0)("Can't create directory %s\n",
					pf->fullname);
			GOTO_ERROR;
		}
dir_ok:
		chown(pf->fullname, pf->ark_uid, pf->ark_gid);
		break;
	case OPB_NAMEDFILE:	/* Try to restore if there is data ??? */
	case OPB_OBJECT:	/* We create a file with object data */
	case OPB_RAWDEV:	/* We create a file with raw data */
	case OPB_REGFILE:
		if (pf->flags & HARD_LINK) {		/* Case of hard link */
			if (link(pf->link_path, pf->fullname) < 0) {
				NL_ERR(0)("Can't link %s to %s\n",
					  pf->fullname, pf->link_path);
				GOTO_ERROR;
			}
			goto ret_oui;
		}
		if (!(pf->ark_mode & ~S_IFMT)) pf->ark_mode = 0644;
		if ((pf->fd = open(pf->fullname,
					(O_WRONLY | O_CREAT | O_TRUNC),
					 pf->ark_mode)) < 0) {
			NL_ERR(0)("Can't create file %s\n", pf->fullname);
			GOTO_ERROR;
		}
		break;
	case OPB_SYMLINK:
		if (!(pf->ark_mode & ~S_IFMT)) pf->ark_mode = 0644;
		if (symlink(pf->link_path, pf->fullname) < 0) {
			if (errno == EEXIST) unlink(pf->fullname);
			if (symlink(pf->link_path, pf->fullname) < 0) {
				NL_ERR(0)("Can't link %s to %s\n",
					  pf->fullname, pf->link_path);
				GOTO_ERROR;
			}
		}
		set_time(pf->fullname, pf->ark_atime, pf->ark_mtime);
		chmod(pf->fullname, pf->ark_mode);
		chown(pf->fullname, pf->ark_uid, pf->ark_gid);
		break;
	default:
		NL_ERR(0)("Unknown file type for file %s\n", pf->fullname);
		GOTO_ERROR;
	}
ret_oui:
	return (OUI);

create_special:
	if (!(pf->ark_mode & ~S_IFMT)) pf->ark_mode = 0644;
	if (mknod(pf->fullname, mode, pf->ark_rdev) < 0) {
		NL_ERR(0)("Can't create special file %s.\n", pf->fullname);
		GOTO_ERROR;
	}
	set_time(pf->fullname, pf->ark_atime, pf->ark_mtime);
	chmod(pf->fullname, pf->ark_mode);
	chown(pf->fullname, pf->ark_uid, pf->ark_gid);
	return (OUI);
error:
	pf->operation = DONT_EXTRACT;
	return (NON);
#endif /* !NLO_WIN32 */
}


int write_file(pf)
struct st_flow *pf;
{
	if (!pf->buff) return(OUI);
	if (pf->fd == INVALID_FD) {
		if (!create_file(pf)) GOTO_ERROR;
	}
	if (pf->flags & COMPRESS_BUFF) uncompress(pf);
	ark_write(pf->fd, pf->buff, pf->buff_size);
	if (!(pf->flags & DONT_FREE_BUFF)) free(pf->buff);
	pf->buff	 = 0;	/* Don't refer to buffer */
	pf->buff_size = 0;
	pf->cur_pc	 = 0;
	return (OUI);
error:
	pf->operation = DONT_EXTRACT;
	if (!(pf->flags & DONT_FREE_BUFF)) free(pf->buff);
	pf->buff	 = 0;
	pf->buff_size = 0;
	pf->cur_pc	 = 0;
	return (NON);
}


static char *getword(line, word, line_end)
char *line;
char *word;
char *line_end;
{
	char *pc1, *pc2;

	for (pc1 = line, pc2 = word;
	     *pc1 && (pc1 < line_end);
	     *pc2 = *pc1, pc1++, pc2++);
	*pc2 = 0;
	if (pc1 == line_end) return (0);
	else return (pc1);
}


static char *getnchars(line, word, line_end, n)
char *line;
char *word;
char *line_end;
int n;
{
	char *pc1, *pc2;
	int i;

	for (pc1 = line, pc2 = word, i = n;
	     i > 0 && *pc1 && (pc1 < line_end);
	     *pc2 = *pc1, pc1++, pc2++, i--);
	*pc2 = 0;
	if (pc1 == line_end) return (0);
	else return (pc1);
}

static int devopenfile(pdev, segnum)
struct st_device *pdev;
int segnum;
{
	char filename[4096];
	sprintf(filename, "%s/seg_%03d", pdev->dvname, segnum);
	if (!(pdev->dvfd = tape_open(filename, O_RDONLY))) GOTO_ERROR;
	return OUI;
error:
	NL_ERR(0)("Can't open file %s\n", filename);
	return NON;
}

struct st_device *devopen(device_name, is_file_tape, block_size)
char *device_name;
int is_file_tape;
int block_size;
{
	ARK_HANDLE fd_in = INVALID_FD;
	int nb_retry = 0;

	pdev = &device_definition;
	memset(pdev, '\0', sizeof(struct st_device));
	pdev->dvname = malloc(strlen(device_name+1));
	strcpy(pdev->dvname, device_name);
	pdev->dvfd = INVALID_FD;
	pdev->dvbufsize = block_size;
	if (block_size) {
		if (!(pdev->dvbuf = malloc(block_size + 65536))) {
			NL_ERR(0)("Can't alloc memory for a device buffer\n");
			myexit(10);
		}
	}
	if (is_file_tape) {
		pdev->dvtype = DV_FILE;
		goto retoui;
	} 
	pdev->dvtype = DV_NORMAL;
	NL_LOG(50)("Openning device %s\n", device_name);
retry:
	if ((fd_in = tape_open(device_name, O_RDONLY)) < 0) {
		if (errno == EAGAIN) {
			if (nb_retry < 5) {
				NL_LOG(40)
					("Resource temporarily unavailable. Retrying...\n");
				nb_retry++;
				sleep(2);
				goto retry;
			}
		}
		NL_ERR(0)("Can't open device %s\n", device_name);
		GOTO_ERROR;
	}
	pdev->dvfd = fd_in;
	pdev->dvatbegseg = 1;
retoui:
	NL_LOG(50)("Device %s open\n", device_name);
	pdev->dvlastread = -1;
	return pdev;
error:
	if (pdev->dvbuf) {
		free(pdev->dvbuf);
		pdev->dvbuf = 0;
	}
	return (struct st_device *)0;
}

int devfree(pdev)
struct st_device *pdev;
{
	if (!pdev) goto retoui;
	if (pdev->dvfd != INVALID_FD) ark_close(pdev->dvfd);
	pdev->dvfd = INVALID_FD;
	if (pdev->dvbuf) free(pdev->dvbuf);
	pdev->dvbuf = 0;
	pdev->dvbufsize = 0;
retoui:
	return OUI;
}


/*
 * read in a buffer: returns the number of bytes read.
 */
int devread(pdev,buf,len)
struct st_device *pdev;
char *buf;
int len;
{
	int nbread = 0;
	int total_nbread = 0;

	if (pdev->dvbufsize) {
		while (len) {
			if (!pdev->dvinbuf) {
				if ((nbread = ark_read(pdev->dvfd, pdev->dvbuf,
						pdev->dvbufsize)) <= 0) {
					goto check_read;
				}
				pdev->dvatbegseg = 0;
				pdev->dvinbuf = nbread;
				pdev->dvbufread = 0;
			}
			if ((pdev->dvbufread + len) >= pdev->dvinbuf) {
				nbread = pdev->dvinbuf - pdev->dvbufread;
				pdev->dvinbuf = 0;
				memcpy(buf + total_nbread, 
						pdev->dvbuf + pdev->dvbufread, 
						nbread);
				pdev->dvbufread = 0;
				total_nbread += nbread;
				len -= nbread;
			} else {
				nbread = len;
				memcpy(buf + total_nbread, 
						pdev->dvbuf + pdev->dvbufread, 
						nbread);
				pdev->dvbufread += nbread;
				len = 0;
				total_nbread += nbread;
			}
		}
		nbread = total_nbread;
		goto retoui;
	}
	if ((nbread = ark_read(pdev->dvfd, buf, len)) > 0) {
		pdev->dvatbegseg = 0;
	}
check_read:
	pdev->dvlastread = nbread;
	if ((nbread < 0) && (pdev->dvcurseg == 0) && (pdev->dvatbegseg == 1)) {
		ark_close(pdev->dvfd);
		pdev->dvfd = INVALID_FD;
		if (pdev->dvtype == DV_FILE) {
			if (!devopenfile(pdev,0)) GOTO_ERROR;
			goto retoui;
		}
		if ((pdev->dvfd = tape_open(pdev->dvname, O_RDONLY))
				== INVALID_FD) {
			NL_LOG(10)("Can't open device \"%s\"\n", pdev->dvname);
		}
	}
retoui:
	return nbread;
error:
	NL_LOG(10)("devread bad end\n");
	return -1;
}

/*
 * read next block : returns the number of bytes read.
 */

static int read_block(pdev, reset)
struct st_device *pdev;
int reset;
{
	char *pc = 0;
	int blk_size;

	if (reset) pdev->dvcurseg = 0;
	blk_size = 0;
	pc	 = (char *)in_blk;
	if (devread(pdev, pc, TAPEPAGESIZE) <= 0) {
		if (errno == ENOENT) errno = 0;
		goto end_of_segment;
	}
	if (need_uint4_chsex) {
		in_blk->blkseqnum = ch_uint4_sex(in_blk->blkseqnum);
		in_blk->blkbkid	  = ch_uint4_sex(in_blk->blkbkid);
	}
	blk_size = (in_blk->blksz & BLKSZMASK) * TAPEPAGESIZE;
	if (blk_size) {
		pc += TAPEPAGESIZE;
		if (devread(pdev, pc, blk_size) < blk_size) {
			if (errno == ENOENT) errno = 0;
			goto end_of_segment;
		}
	}
	blk_size += TAPEPAGESIZE;
	return (blk_size);

end_of_segment:
	NL_LOG(10)("read_block: Got end of segment '%d'\n", pdev->dvcurseg);
	pdev->dvcurseg++;
	return (0);
}


void get_string_vals(pf)
struct st_flow *pf;
{
	char str[STR_SIZE];

	NL_LOG(50)("[%d] Prop '%s'='%s'\n",
		   pf->fnum, pf->prop, pf->value);

	if (!strcmp(pf->prop, "s_abxvz_ftype")) {
		if (!strcmp(pf->value, "S_IFOBJ"))	  pf->type = OPB_OBJECT;
		else if(!strcmp(pf->value, "S_IFDEV")) pf->type = OPB_RAWDEV;
		else if(!strcmp(pf->value, "S_IFDIR")) pf->type = OPB_DIR;
		else if(!strcmp(pf->value, "S_IFCHR")) pf->type = OPB_CHRDEV;
		else if(!strcmp(pf->value, "S_IFBLK")) pf->type = OPB_BLKDEV;
		else if(!strcmp(pf->value, "S_IFREG")) pf->type = OPB_REGFILE;
		else if(!strcmp(pf->value, "S_IFIFO")) pf->type = OPB_FIFO;
		else if(!strcmp(pf->value, "S_IFNAM")) pf->type = OPB_NAMEDFILE;
		else if(!strcmp(pf->value, "S_IFLNK")) pf->type = OPB_SYMLINK;
		else {
			NL_ERR(0)("Unknown file type: %s\n", pf->value);
			pf->type = OPB_UNKNOWN;
		}
	} else if (!strcmp(pf->prop, "s_host")) {
		strcpy(pf->host, pf->value);
		*(pf->volume) = 0;
		*(pf->location) = 0;
		if ((global_operation == ADMIN_INFO) &&
		    pf->slave_version && *pf->slave_version) {
			INFO(0)("[%d] Got file(s) from slave '%s' version %s\n",
				pf->fnum, pf->host, pf->slave_version);
			*pf->slave_version = 0;
		}
	} else if (!strcmp(pf->prop, "s_volum")) {
		strcpy(pf->volume, pf->value);
	} else if (!strcmp(pf->prop, "s_pluga")) {
		strcpy(pf->plugin, pf->value);
	} else if (!strcmp(pf->prop, "s_locat")) {
		strcpy(pf->location, pf->value);
	} else if (!strcmp(pf->prop, "s_cmpnt")) {
		strcpy(pf->component, pf->value);
	} else if (!strcmp(pf->prop, "s_fslnk")) {
		strcpy(pf->link_path, pf->value);
	} else if (!strcmp(pf->prop, "s_flink")) {
		sprintf(str, "%s/file%s", pf->host, pf->value);
		strcpy(pf->link_path, str);
	} else if (!strcmp(pf->prop, "s_typeobj")) {
		strcpy(pf->obj_type, pf->value);
	}
}


void get_mode_str(pf, str)
struct st_flow *pf;
char *str;
{
	mode_t mode;
	char *pc;

	pc = str;
	switch(pf->type) {
	case OPB_BLKDEV:  *pc = 'b'; break;
	case OPB_CHRDEV:  *pc = 'c'; break;
	case OPB_FIFO:	  *pc = 'p'; break;
	case OPB_DIR:	  *pc = 'd'; break;
	case OPB_SYMLINK: *pc = 'l'; break;
	case OPB_OBJECT:  *pc = 'O'; break;
	case OPB_RAWDEV:  *pc = 'R'; break;
	default:	  *pc = '-'; break;
	}
	if (pf->type == OPB_DIR && !(pf->ark_mode & ~S_IFMT)) pf->ark_mode = 0755;
	else if(!(pf->ark_mode & ~S_IFMT)) pf->ark_mode = 0644;
	pc++;

	mode = pf->ark_mode;
	if (mode & S_IREAD) *pc = 'r';
	else *pc = '-';
	pc++;
	if (mode & S_IWRITE) *pc = 'w';
	else *pc = '-';
	pc++;
	if ((mode & S_ISUID) && (mode & S_IEXEC)) *pc = 's';
	else if (mode & S_IEXEC) *pc = 'x';
	else if (mode & S_ISUID) *pc = 'S';
	else *pc = '-';
	pc++;

	if (mode & (S_IREAD>>3)) *pc = 'r';
	else *pc = '-';
	pc++;
	if (mode & (S_IWRITE>>3)) *pc = 'w';
	else *pc = '-';
	pc++;
	if ((mode & S_ISGID) && (mode & (S_IEXEC>>3))) *pc = 's';
	else if (mode & (S_IEXEC>>3)) *pc = 'x';
	else if (mode & S_ISGID) *pc = 'l';
	else *pc = '-';
	pc++;

	if (mode & (S_IREAD>>6)) *pc = 'r';
	else *pc = '-';
	pc++;
	if (mode & (S_IWRITE>>6)) *pc = 'w';
	else *pc = '-';
	pc++;
	if ((mode & S_ISVTX) && (mode & (S_IEXEC>>6))) *pc = 't';
	else if (mode & (S_IEXEC>>6)) *pc = 'x';
	else if (mode & S_ISVTX) *pc = 'T';
	else *pc = '-';
	pc++;
	*pc = 0;
}


void get_num_vals(pf)
struct st_flow *pf;
{
	char *pc;
	char mode_str[STR_SIZE];
	char file_str[STR_SIZE];
	int num;

	num = atoi(pf->value);
	NL_LOG(70)("[%d] Prop '%s'='%d'\n", pf->fnum, pf->prop, num);

	if (!strcmp(pf->prop, "n_dev"))	 pf->ark_dev	  = num;
	else if (!strcmp(pf->prop, "n_ino"))	 pf->ark_ino	  = num;
	else if (!strcmp(pf->prop, "n_mode"))	 pf->ark_mode  = num;
#ifndef NLO_WIN32
	else if (!strcmp(pf->prop, "n_nlink")) pf->ark_nlink = num;
	else if (!strcmp(pf->prop, "n_uid"))	 pf->ark_uid	  = num;
	else if (!strcmp(pf->prop, "n_gid"))	 pf->ark_gid	  = num;
#endif /* !NLO_WIN32 */
	else if (!strcmp(pf->prop, "n_rdev"))	 pf->ark_rdev  = num;
	else if (!strcmp(pf->prop, "n_fsize")) pf->ark_size  = num;
	else if (!strcmp(pf->prop, "n_atime")) pf->ark_atime = num;
	else if (!strcmp(pf->prop, "n_mtime")) pf->ark_mtime = num;
	else if (!strcmp(pf->prop, "n_ctime")) pf->ark_ctime = num;
	else if (!strcmp(pf->prop, "n_cksum")) {
		if (pf->operation == NO_OPERATION) {
			/* No data comming with this file */
			check_operation(pf);
			if (pf->type == OPB_REGFILE) {
#ifndef NLO_WIN32
				if (pf->ark_nlink > 1) {
					pf->flags |= HARD_LINK;
				}
#endif /* !NLO_WIN32 */
			}
			if (pf->operation == EXTRACT_FILE) create_file(pf);
		}
		else if (pf->operation == EXTRACT_FILE) {
			NL_LOG(10)("[%d] x %s\n", pf->fnum, pf->fullname);
		}
		else if (pf->operation == CONSULT_LIST) {
			get_mode_str(pf, mode_str);
			pc = ctime(&pf->ark_mtime);
			if (pf->type == OPB_SYMLINK) {
				sprintf(file_str, "%s symbolic link to %s",
					pf->fullname, pf->link_path);
			} else {
				if (pf->flags & HARD_LINK) {
					sprintf(file_str, "%s linked to %s",
						pf->fullname, pf->link_path);
				} else sprintf(file_str,"%s", pf->fullname);
			}
			if (!nb_path || match(pf->fullname)) {
			NL_LOG(10)("[%d] %s %4d/%-4d%7ld %-12.12s %-4.4s %s\n",
				pf->fnum, mode_str,
#ifdef NLO_WIN32
				0, 0,
#else
				pf->ark_uid, pf->ark_gid,
#endif /* NLO_WIN32 */
				(long)pf->ark_size, pc+4,
				pc+20, file_str);
			}
		}
		if (pf->operation == DONT_EXTRACT) {
			NL_LOG(30)("[%d] s %s\n", pf->fnum, pf->fullname);
		}
		end_of_file(pf);
	}
}


void end_of_file(pf)
struct st_flow *pf;
{
	UINT4 ui1, ui2;
	char *p1, *p2, *p3, *p4, *p5, *p6, *p7, *p8, *p9, *p10, *p11, *p12;
	int n1;

	if (pf->operation == EXTRACT_FILE) {
		if (pf->buff) write_file(pf);
		if (pf->fd >= 0) {
			ark_close(pf->fd);
			pf->fd = INVALID_FD;
			set_time(pf->fullname, pf->ark_atime,
				 pf->ark_mtime);
			if (!(pf->ark_mode & ~S_IFMT)) pf->ark_mode = 0644;
			chmod(pf->fullname, pf->ark_mode);
#ifndef NLO_WIN32
			chown(pf->fullname, pf->ark_uid, pf->ark_gid);
#endif /* !NLO_WIN32 */
		}
	}

	/* Initialise flow vars */
	p1 = pf->prop;
	p2 = pf->value;
	p3 = pf->realsize;
	p4 = pf->fullname;
	p5 = pf->host;
	p6 = pf->volume;
	p7 = pf->location;
	p8 = pf->component;
	p9 = pf->link_path;
	p10 = pf->obj_type;
	p11 = pf->slave_version;
	p12 = pf->plugin;
	if (p1) *p1 = *p2 = *p3 = *p4 = *p8 = *p9 = *p10 = *p11 = *p12 = 0;

	n1 = pf->fnum;
	ui1 = pf->bkid;
	ui2 = pf->blknum;

	memset(pf, '\0', sizeof(struct st_flow));

	pf->prop = p1;
	pf->value = p2;
	pf->realsize	= p3;
	pf->fullname = p4;
	pf->host = p5;
	pf->volume = p6;
	pf->location = p7;
	pf->component = p8;
	pf->link_path = p9;
	pf->obj_type = p10;
	pf->slave_version = p11;
	pf->plugin = p12;

	pf->fnum = n1;
	pf->bkid = ui1;
	pf->blknum = ui2;

	pf->state = NORMAL_STATE;
	pf->operation = NO_OPERATION;
	pf->fd = INVALID_FD;
}


char *perform_empty_block(pf, pc)
struct st_flow *pf;
char *pc;
{
	int realsize;

	realsize = atoi(pf->realsize);

	NL_LOG(70)("[%d] Empty block prop '%s' (%d)\n",
		   pf->fnum, pf->prop, realsize);
	if (pf->operation == EXTRACT_FILE) {
		if (pf->buff) {
			if (!write_file(pf)) goto ret_oui;
		} else {
			if (pf->fd == INVALID_FD) {
				if (!create_file(pf)) goto ret_oui;
			}
		}
		if (realsize) {
			if (ark_lseek(pf->fd, (off_t) (realsize - 1),
					SEEK_CUR) < 0) {
				NL_ERR(0)("Can't seek for free block\n");
				myexit(4);
			}
			ark_write(pf->fd, "\0", 1);
		}
		goto ret_oui;
	}
ret_oui:
	/* If don't extract forget this empty block */
	pf->state = NORMAL_STATE;
	return(pc);
}


char *get_bin_vals(pf, pc, max_pc)
struct st_flow *pf;
char *pc;
char *max_pc;
{
	char *pc1;
	int data_len, realsize, max_len, nb_bytes;

	data_len = pf->bin_rest;
	realsize = atoi(pf->realsize);

	if (pf->operation == NO_OPERATION) check_operation(pf);

	if (!strcmp(pf->prop, "b_9f")) return (perform_empty_block(pf, pc));

	if (pf->operation == EXTRACT_FILE) { /* File data */
		if (strlen(pf->prop) == 4) { /* Block mode */
			if (pf->state == INCOMPLETE_BIN) {
				NL_LOG(70)("[%d] Continuing Binary prop '%s' (%d)\n",
					   pf->fnum, pf->prop, data_len);
			} else {
				if (pf->buff) {
					if (!write_file(pf)) {
						pf->operation = DONT_EXTRACT;
						goto skip_data;
					}
				}
				if (realsize &&
				    ((hextoi(pf->prop + 2) & COMPRESS_MASK) > 0)) {
					pf->flags |= COMPRESS_BUFF;
					NL_LOG(70)
					  ("[%d] %s Compress Bin prop '%s' (size=%d) (realsize=%d)\n",
					   pf->fnum, compress_type(pf->prop),
					   pf->prop, data_len, realsize);
				} else {
					NL_LOG(70)("[%d] Binary prop '%s' (size=%d)\n",
						   pf->fnum, pf->prop, data_len);
				}
				if (!(pf->buff = malloc(data_len))) {
					NL_ERR(0)("Can't alloc memory for a file buffer\n");
					myexit(4);
				}
				pf->buff_size = data_len;
				pf->cur_pc = pf->buff;
			}
			for (pc1 = pf->cur_pc;
			     (pc < max_pc) && (data_len > 0);
			     *pc1 = *pc, pc++, pc1++, data_len--);
			if (data_len) {
				pf->cur_pc = pc1;
				pf->state = INCOMPLETE_BIN;
				pf->bin_rest = data_len;
				goto unexp_eob;
			} else write_file(pf);
			goto ret_oui;
		} else if (!strcmp(pf->prop, "b_fsiz")) {
			if (pf->fd == INVALID_FD) {
				if (!create_file(pf)) {
					pf->operation = DONT_EXTRACT;
					goto skip_data;
				}
			}
			max_len = max_pc - pc;
			nb_bytes = (max_len > data_len) ? data_len : max_len;

			ark_write(pf->fd, pc, nb_bytes);

			data_len -= nb_bytes;
			pc += nb_bytes;

			if (data_len) {
				pf->state = INCOMPLETE_BIN;
				pf->bin_rest = data_len;
				goto unexp_eob;
			}
		} else goto skip_data;
	} else {
		/* Skip those binary data */
		if (pf->state == INCOMPLETE_BIN) {
			NL_LOG(70)("[%d] Continuing to skip Binary prop '%s' (%d)\n",
				   pf->fnum, pf->prop, data_len);
		} else {
skip_data:
			if ((strlen(pf->prop) == 4) && realsize &&
			   ((hextoi(pf->prop + 2) & COMPRESS_MASK) > 0)) {
				pf->flags |= COMPRESS_BUFF;
				NL_LOG(70)
				  ("[%d] Skipping %s compress bin prop '%s' (size=%d) (realsize=%d)\n",
				   pf->fnum, compress_type(pf->prop),
				   pf->prop, data_len, realsize);
			} else {
				NL_LOG(70)("[%d] Skipping binary prop '%s' (size=%d)\n",
					   pf->fnum, pf->prop, data_len);
			}
		}
		max_len = max_pc - pc;
		nb_bytes = (max_len > data_len) ? data_len : max_len;

		ark_write(pf->fd, pc, nb_bytes);

		data_len -= nb_bytes;
		pc += nb_bytes;

		if (data_len) {
			pf->state = INCOMPLETE_BIN;
			pf->bin_rest = data_len;
			goto unexp_eob;
		}
	}
ret_oui:
	pf->state = NORMAL_STATE;
	return (pc);
unexp_eob:
	return (0);
}


char *parse_value(pf, pc, max_pc)
struct st_flow *pf;
char *pc;
char *max_pc;
{
	pf->state = INCOMPLETE_VAL;
	switch (*pf->prop) {
	case 's':	/* String */
		GETWORD(pc, pf->value, max_pc);
		get_string_vals(pf);
		break;
	case 'n':	/* Number */
		GETWORD(pc, pf->value, max_pc);
		get_num_vals(pf);
		break;
	case 'b':
		GETWORD(pc, pf->value, max_pc);	/* Block size */
		pf->bin_rest = atoi(pf->value);
		if (strlen(pf->prop) == 4) {
			pf->state = INCOMPLETE_REALSZ; /* Real size */
			GETNCHARS(pc, pf->realsize, max_pc, 5);
			pf->bin_rest -= 5;
		}
		if (!(pc = get_bin_vals(pf, pc, max_pc))) goto unexp_eob;
		break;
	}
	pf->state = NORMAL_STATE;
	return (pc);
unexp_eob:
	return (0);
}


char *complete_missing(pf, pc, max_pc)
struct st_flow *pf;
char *pc;
char *max_pc;
{
	char str[STR_SIZE];
	int n;

	if (pf->state == INCOMPLETE_PROP) {
		GETWORD(pc, str, max_pc);
		strcat(pf->prop, str);
		if (*pf->prop == 'b') {
			GETWORD(pc, pf->value, max_pc); /* Block size */
			pf->bin_rest = atoi(pf->value);
			if (strlen(pf->prop) == 4) {
				pf->state = INCOMPLETE_REALSZ;
				GETNCHARS(pc, pf->realsize, max_pc, 5); /* Real size */
				pf->bin_rest -= 5;
			}
			return (get_bin_vals(pf, pc, max_pc));
		}
		if (!(pc = parse_value(pf, pc, max_pc))) goto unexp_eob;
		return (pc);
	}
	if (pf->state == INCOMPLETE_VAL) {
		GETWORD(pc, str, max_pc);
		strcat(pf->value, str);
		if (*pf->prop == 'b') {
			pf->bin_rest = atoi(pf->value);
			if (strlen(pf->prop) == 4) {
				GETNCHARS(pc, pf->realsize, max_pc, 5); /* Real size */
				pf->bin_rest -= 5;
			}
			return (get_bin_vals(pf, pc, max_pc));
		}
		strcpy(str, pf->value);
		(void) parse_value(pf, str, &str[STR_SIZE]);
		return (pc);
	}
	if (pf->state == INCOMPLETE_REALSZ) {
		n = strlen(pf->realsize);
		GETNCHARS(pc, str, max_pc, 5-n);
		strcat(pf->realsize, str);
		pf->bin_rest = atoi(pf->value) - 5;
		return (get_bin_vals(pf, pc, max_pc));
	}
	if (pf->state == INCOMPLETE_BIN) {
		return (get_bin_vals(pf, pc, max_pc));
	}
	return (pc);
unexp_eob:
	return (NON);
}


int parse_block(pf, pblk, blk_size)
struct st_flow *pf;
struct stapblk *pblk;
int blk_size;
{
	char *pc, *max_pc;

	max_pc = ((char *)pblk) + blk_size;

	if (pf->state == NORMAL_STATE) {
		pc =((char *)pblk) + ((pblk->blkproffh << 8) | pblk->blkproffl);
	} else {
		pc = ((char *)pblk) + HEADER_SIZE;
		if (!(pc = complete_missing(pf, pc, max_pc))) goto unexp_eob;
	}

	pf->state = NORMAL_STATE;
	while (*pc && (pc < max_pc)) {
		pf->state = INCOMPLETE_PROP;
		GETWORD(pc, pf->prop, max_pc);
		if (!strcmp(pf->prop, OPB_SERV_VERSION)) {
			GETWORD(pc, pf->value, max_pc);
			strcpy(pf->slave_version, pf->value);
			pf->state = NORMAL_STATE;
			continue;
		}
		if (!(pc = parse_value(pf, pc, max_pc))) goto unexp_eob;
	}
	return (OUI);
unexp_eob:
	NL_LOG(40)("[%d] We need another block to complete.\n", pf->fnum);
	return (OUI);
}


int read_label(pdev)
struct st_device *pdev;
{
	int offset;
	char *pc, *max_pc;
	char prop[STR_SIZE];
	char value[STR_SIZE];
	int label_size = 0;
	int num;
	struct stapblk *pblk=0;
	char tapeid[STR_SIZE];
	char author[STR_SIZE];
	char label_name[STR_SIZE];

	if (!(label_size = read_block(pdev, 1))) GOTO_ERROR;
	pblk = in_blk;
	if (pblk->blkseqnum != O3MAGICTAPE) {
		if (ch_uint4_sex(pblk->blkseqnum) != O3MAGICTAPE) {
			NL_ERR(0)("This tape is not an ARKEIA tape.\n");
			GOTO_ERROR;
		}
	}
	offset = ((pblk->blkproffh << 8) | pblk->blkproffl);
	pc = ((char *)pblk) + offset;
	max_pc = ((char *)pblk) + label_size;

	while (*pc && (pc < max_pc)) {
		GETWORD(pc, prop, max_pc);
		if (!strcmp(prop, OPB_VERSION)) {
			GETWORD(pc, value, max_pc);
			INFO(0)("Server version %s\n", value);
			continue;
		}
		switch (*prop) {
		case 's':	/* String */
			GETWORD(pc, value, max_pc);
			NL_LOG(60)("Prop '%s'='%s'\n", prop, value);
			if (!strcmp(prop, "s_lnam")) strcpy(label_name, value);
			else if (!strcmp(prop, "s_auth")) strcpy(author, value);
			break;
		case 'n':
			GETWORD(pc, value, max_pc);
			num = atoi(value);
			NL_LOG(60)("Prop '%s'='%d'\n", prop, num);
			if (!strcmp(prop, "n_tapeid")) strcpy(tapeid, value);
			break;
		}
	}
	if (global_operation == ADMIN_INFO) {
		INFO(0)("Reading tape '%s' created by %s (tapeid=%s)\n",
			label_name, author, tapeid);
	} else if(global_operation == EXTRACT_FILE) {
		NL_LOG(10)("Extracting data from tape '%s' created by %s\n",
			   label_name, author);
	} else {
		NL_LOG(10)("Consulting data from tape '%s' created by %s\n",
			   label_name, author);
	}
	while((read_block(pdev, 0))>0);

	return(label_size);
error:
	NL_ERR(0)("Can't read tape label.\n");
	return (0);
unexp_eob:
	NL_ERR(0)("Found unexpected end of block.\n");
	return (0);
}


#define MBUFSIZE	4096		/* 4K buffer */

int read_tape(device_name, label_check, block_size)
char *device_name;
int label_check;
int block_size;
{
	char *buf;
	char temp_file[STR_SIZE];
	char *temp_dir = 0;
	int blk_size;
	UINT4 blknum;
	int flow_num;
	int file_tape = 0, delete = NON;
	int n;
	int nb_null;
	ARK_HANDLE tmp_fd = INVALID_FD;
	struct st_flow *pf;
	struct stat stbuf;

	if(global_operation == ADMIN_INFO) {
		INFO(0)("Using readarkeia version '%s'\n", readarkeia_version);
	}

	NL_LOG(10)("Reading device '%s'\n", device_name);

	delete = NON;
	if(!strcmp(device_name, "-")) {
		if (!(temp_dir = getenv("ARK_TMP_DIR"))) temp_dir = TEMP_DIR;

		sprintf(temp_file, "%s/readarkeia_%d", temp_dir,getpid());
		NL_LOG(10)("Using temporary file '%s'\n", temp_file);

#ifdef NLO_WIN32
		tmp_fd = CreateFile(temp_file, GENERIC_WRITE, 0, 0,
			CREATE_ALWAYS,
			FILE_FLAG_BACKUP_SEMANTICS | FILE_FLAG_SEQUENTIAL_SCAN,
			0);
#else
		tmp_fd = open(temp_file,
			O_RDWR | O_CREAT | O_TRUNC | O_EXCL, 0644);
#endif /* NLO_WIN32 */
		if (tmp_fd == INVALID_FD) {
			NL_ERR(0)("readarkeia: temporary file %s\n", temp_file);
			myexit(2);
		}
		buf = malloc(MBUFSIZE);
		if (!buf) {
			NL_ERR(0)("tar: Out of memory\n");
			myexit(2);
		}
		while ((n = read(0, buf, MBUFSIZE)) > 0) {
			ark_write(tmp_fd, buf, n);
		}
		free(buf);
		ark_close(tmp_fd);
		tmp_fd = INVALID_FD;
		device_name = temp_file;
		delete = OUI;
	} else {
		if (stat(device_name, &stbuf) == 0 && S_ISDIR(stbuf.st_mode)) {
			file_tape = 1;
		}
	}
	if (!(pdev = devopen(device_name, file_tape, block_size))) {
		NL_ERR(0)("Can't open device %s\n", device_name);
		GOTO_ERROR;
	}
	if (!(in_blk = (struct stapblk *)malloc(64 * 1024))) {
		NL_ERR(0)("Can't alloc memory for label blk\n");
		GOTO_ERROR;
	}
	if (pdev->dvtype==DV_FILE) {
		if (!devopenfile(pdev, 0)) GOTO_ERROR;
	}
	if (label_check && !read_label(pdev)) GOTO_ERROR;

	nb_null = 0;
	while (1) {
		if (nb_null > 3) {
			NL_LOG(40)("End of tape.\n");
			goto ret_oui;
		}
		if (!(blk_size = read_block(pdev, 0))) {
			/* End of segment */
			if (pdev->dvtype == DV_FILE){
				ark_close(pdev->dvfd);
				pdev->dvfd = INVALID_FD;
				if (!devopenfile(pdev, pdev->dvcurseg)) {
					nb_null = 4;
					continue;
				}
				NL_LOG(20) ("Reading File segment '%d'\n",
					pdev->dvcurseg);
			}
			nb_null++;
			continue;
		}
		nb_null = 0;
		flow_num = COMP2FLOW(in_blk->blkseqnum);
		if (flow_num == F_ST) {
			NL_LOG(70)("Got Strategy process block. skipping...\n");
			continue;
		}
		if (in_blk->blksz & DB_BLOCK) {
			NL_LOG(40)("Got database block. skipping...\n");
			continue;
		}

		/* It's a data block */
		if ((flow_num < 0) || (flow_num >= MAXFLOWS)) {
			NL_ERR(0)
				("Found a block with a flow number higher than %d: %d\n",
				 MAXFLOWS, flow_num);
			myexit(9);
		}
		blknum = (in_blk->blkseqnum & 0xffffff);
		pf = &flow_tab[flow_num-1];
		if (blknum > (pf->blknum+1)){ /* It's another backup */
			if ((pf->state != NORMAL_STATE) ||
					(pf->buff)) {
				NL_LOG(10)
					("[%d] Warning file %s has been damaged.\n",
					 pf->fnum, pf->fullname);
				if (pf->operation == EXTRACT_FILE) {
					NL_LOG(10)("[%d] Extracting new file.\n",
							pf->fnum);
				}
			}
			end_of_file(pf);
		} else {
			if (blknum <= pf->blknum && (in_blk->blkbkid == pf->bkid)) { /* Got duplicate block */
				NL_LOG(30)
					("Got an already read block %d for flow %d. skipping...\n",
					 blknum, pf->fnum);
				continue;
			}
		}
		NL_LOG(60)("[%d] Start of Block %d, size %d, bkid %d.\n",
				pf->fnum, blknum, blk_size, in_blk->blkbkid);

		/* Allocate strings needed by this flow is they are not already */
		ALLOC_IF_NOT(pf->prop);
		ALLOC_IF_NOT(pf->value);
		ALLOC_IF_NOT(pf->realsize);
		ALLOC_IF_NOT(pf->fullname);
		ALLOC_IF_NOT(pf->host);
		ALLOC_IF_NOT(pf->plugin);
		ALLOC_IF_NOT(pf->volume);
		ALLOC_IF_NOT(pf->location);
		ALLOC_IF_NOT(pf->component);
		ALLOC_IF_NOT(pf->link_path);
		ALLOC_IF_NOT(pf->obj_type);
		ALLOC_IF_NOT(pf->slave_version);

		if (pf->bkid && (pf->bkid != in_blk->blkbkid)) {
			NL_ERR(0)
				("Got another backup id for flow %d: %d --> %d\n",
				 pf->fnum, pf->bkid, in_blk->blkbkid);
		}
		pf->blknum = blknum;
		if (!parse_block(pf, in_blk, blk_size)) GOTO_ERROR;
	}
ret_oui:
	devfree(pdev);
	if (delete) {
#ifdef NLO_WIN32
		DeleteFile(device_name);
#else
		unlink(device_name);
#endif /* NLO_WIN32 */
	}
	return (OUI);
error:
	devfree(pdev);
	if (delete) {
#ifdef NLO_WIN32
		DeleteFile(device_name);
#else
		unlink(device_name);
#endif /* NLO_WIN32 */
	}
	return (NON);
}


int main(argc, argv)
int argc;
char **argv;
{
	char *pc;
	char c;
	char device_name[STR_SIZE];
	int i;
	int inc;
	int label_check;
	int block_size = 0;

	argv++;
	if (!--argc) usage();
	if (argv[0][0] == '-') {
		switch (argv[0][1]) {
		case 'l':
			nl_loglevel = atoi(&argv[0][2]);
			break;
		}
		argc--;
		argv++;
	}
	if (!argc) usage();
	inc = 1;
	label_check = OUI;
	for (i = 0; argv[0][i]; i++) {
		switch (argv[0][i]) {
		case 'i':
			if (global_operation != NO_OPERATION) usage();
			global_operation = ADMIN_INFO;
			break;
		case 'f':
			if (argc < inc +1) usage();
			strcpy(device_name, argv[inc]);
			inc++;
			break;
		case 'm':
			label_check = NON;
			break;
		case 't':
			if (global_operation != NO_OPERATION) usage();
			global_operation = CONSULT_LIST;
			break;
		case 'v':
			if (!nl_loglevel) nl_loglevel = 10;
				/* First level of verbose */
			break;
		case 'x':
			if (global_operation != NO_OPERATION) usage();
			global_operation = EXTRACT_FILE;
			break;
		case 'u':
			gl_update_mode = OUI;
			break;
		case 'b':
			if (argc < (inc +1)) usage();
			block_size = atoi(argv[inc]);
			inc++;
			break;
		default:
			usage();
		}
	}
	argc -= inc;
	argv += inc;
	while (argc) {
		i = strlen(argv[0]);
		pc = malloc(i+1);
		strcpy(pc, argv[0]);
		extract_path[nb_path] = pc;
		nb_path++;
		argc--;
		argv++;
	}
	if (!*device_name) usage();
	INFO(0)("Device name = %s\n", device_name);
	for (i = 0; i < MAXFLOWS; i++) {
		memset((char *)(&flow_tab[i]), '\0', sizeof(struct st_flow));
		flow_tab[i].fnum = i;
		flow_tab[i].state = NORMAL_STATE;
		flow_tab[i].operation = NO_OPERATION;
		flow_tab[i].fd = INVALID_FD;
	}
	while (1) {
		if (!read_tape(device_name, label_check, block_size)) GOTO_ERROR;
		label_check = OUI;
		if (!strcmp(device_name, "-")) goto ret_oui;
		printf("Found end of tape. ");
		printf("Please insert next tape to continue\nand press return or ");
		printf("press 'q' to quit.\n<PRESS ONE KEY>\n");
		fflush(stdout);
		c = getchar();
		if ((c == 'q') || (c == 'Q')) break;
	}
ret_oui:
	myexit(0);
	return (0);
error:
	myexit(3);
	return (3);
}


void myexit(i)
int i;
{
	if (fd_in != INVALID_FD) {
		ark_close(fd_in);
		fd_in = INVALID_FD;
	}
	exit(i);
}


void usage()
{
	fprintf(stderr, "USAGE (V%s):\n\t", readarkeia_version);
	fprintf(stderr,
"readarkeia [-l<loglevel>] x|t|i [mbfv] [device|-] [filename1 filename2 ...]\n");
	fprintf(stderr,"\nMain operation mode:\n");
	fprintf(stderr,"\tx: extract files from an archive\n");
	fprintf(stderr,"\tt: list the contents of an archive\n");
	fprintf(stderr,"\ti: display backup informations\n");
	fprintf(stderr,"\nEnvironment variable:\n");
	fprintf(stderr,
"\tARK_TMP_DIR: change the temporary directory (default is %s)\n\n", TEMP_DIR);
	myexit(2);
}



int hextoi(str)
char *str;
{
	register char *pc;
	int i;

	for (i = 0, pc = str; ; pc++) {
		switch (*pc) {
		case 0:
		case '\t':
		case ' ':
			continue;
		}
		break;
	}
	for (; ; pc++) {
		if ((*pc >= '0') && (*pc <= '9')) i = i * 16 + (*pc - '0');
		else if ((*pc >= 'a') && (*pc <= 'f')) i = i * 16 + (*pc - 'a' + 10);
		else if ((*pc >= 'A') && (*pc <= 'F')) i = i * 16 + (*pc - 'A' + 10);
		else break;
	}
	return (i);
}

/* UNCOMPRESS LAYER*/

static char *pwrkmem;
static ULONG wrklen = 0;

typedef void (*void_funcp)();

#define fast_copy(src,dest,len) memcpy(dest,src,len)

#define U(X) ((ULONG) X)

#define FLAG_BYTES	   4
#define FLAG_COPY	   1


char *compress_type(prop)
char *prop;
{
	int mask;

	mask = hextoi(prop + 2);
	if ((mask & COMPRESS_MASK) == LZ1_COMPRESS) return ("lz1");
	else if ((mask & COMPRESS_MASK) == LZ3_3_COMPRESS) return ("lz3_3");
	else return("??");
}

/*Based on lzrw1-a.c by Ross Williams */

ULONG lz1_memory = U(U(sizeof(long))*U(4096)+U(3));

static void lz1_decompress(wrk_mem, p_src_first, src_len, p_dst_first, p_dst_len)
UBYTE *p_src_first, *p_dst_first;
ULONG src_len, *p_dst_len;
UBYTE *wrk_mem;
{
	register UBYTE *p_src = p_src_first+FLAG_BYTES, *p_dst = p_dst_first;
	UBYTE *p_src_post = p_src_first+src_len;
	UBYTE *p_src_max16 = p_src_first+src_len-(16*2);
	register ULONG control = 1; /* Put 1 in variable control */

	if (*p_src_first == FLAG_COPY) {
		fast_copy(p_src_first + FLAG_BYTES, p_dst_first, src_len - FLAG_BYTES);
		*p_dst_len = src_len-FLAG_BYTES;
		return ;
	}
	while (p_src!=p_src_post) {
		register UWORD unroll;

		if (control == 1) {
			control = 0x10000|*p_src++;
			control |= (*p_src++) << 8;
		}
		unroll =  p_src <= p_src_max16 ? 16 : 1;
		while (unroll--) {
			if (control&1) {
				register UWORD lenmt;
				register UBYTE *p;

				lenmt = *p_src++;
				p = p_dst - (int)(((lenmt&0xF0) << 4) | *p_src++);
				*p_dst++ = *p++;
				*p_dst++ = *p++;
				*p_dst++ = *p++;
				lenmt &= 0xF;
				while (lenmt--) *p_dst++ = *p++;
			} else *p_dst++ = *p_src++;
			control >>= 1;
		}
	}
	*p_dst_len = p_dst-p_dst_first;
}

/* Based on lzrw3-a.c by Ross Williams */

#define SIZE_P_BYTE	(U(sizeof(UBYTE *)))
#define ALIGNMENT_FUDGE (U(16))
#define MEM_REQ		(U(4096)*(SIZE_P_BYTE) + ALIGNMENT_FUDGE)

ULONG lz3_3_memory = MEM_REQ;


#define HASH_TABLE_DEPTH_BITS		(3)
#define HASH_TABLE_LENGTH		(4096)
#define MAX_CMP_GROUP			(2+16*2)
#define MAX_RAW_GROUP			(16*MAX_RAW_ITEM)
#define MAX_RAW_ITEM			(18)
#define ULONG_ALIGN_UP(X)		((((ULONG)X)+3)&~3)

#define PARTITION_LENGTH_BITS		(12-HASH_TABLE_DEPTH_BITS)
#define PARTITION_LENGTH		(1<<PARTITION_LENGTH_BITS)
#define HASH_TABLE_DEPTH		(1<<HASH_TABLE_DEPTH_BITS )
#define HASH_MASK			(PARTITION_LENGTH-1)
#define DEPTH_MASK			(HASH_TABLE_DEPTH-1)

#define START_STRING_18			((UBYTE *) "123456789012345678")

#define HASH(PTR) \
( \
 (((40543*(((UINT)(*(PTR))<<8)^((UINT)(*((PTR)+1))<<4)^(UINT)(*((PTR)+2))))>>4) & HASH_MASK) \
 << HASH_TABLE_DEPTH_BITS \
 )

#define UPDATE_P(P_BASE,NEWPTR) \
{(P_BASE)[cycle++] = (NEWPTR); cycle &= DEPTH_MASK;}

#define UPDATE_I(I_BASE,NEWPTR) \
{hash[(I_BASE)+cycle++] = (NEWPTR); cycle &= DEPTH_MASK;}

#define ANY_HASH_INDEX (0)


#define ZJ *p_h++ = START_STRING_18

static void lz3_3_decompress(p_wrk_mem, p_src_first, src_len, p_dst_first, p_dst_len)
UBYTE *p_wrk_mem;
UBYTE *p_src_first;
ULONG  src_len;
UBYTE *p_dst_first;
ULONG *p_dst_len;
{
	register UBYTE *p_src = p_src_first+FLAG_BYTES;
	register UBYTE *p_dst = p_dst_first;
	UBYTE *p_src_post  = p_src_first + src_len;
	UBYTE *p_src_max16 = p_src_first + src_len - (MAX_CMP_GROUP - 2);
	UBYTE **hash = (UBYTE **) ULONG_ALIGN_UP(p_wrk_mem);
	register ULONG control = 1;
	register UCARD literals = 0;
	UCARD cycle = 0;

	if (*p_src_first == FLAG_COPY) {
		fast_copy(p_src_first + FLAG_BYTES, p_dst_first, src_len - FLAG_BYTES);
		*p_dst_len = src_len - FLAG_BYTES;
		return ;
	}
	{
		UCARD i;
		UBYTE **p_h = hash;
		for (i = 0; i < 256; i++) {
			ZJ; ZJ; ZJ; ZJ;
			ZJ; ZJ; ZJ; ZJ;
			ZJ; ZJ; ZJ; ZJ;
			ZJ; ZJ; ZJ; ZJ;
		}
	}
	while (p_src != p_src_post) {
		register UCARD unroll;

		if (control == 1) {
			control = 0x10000| * p_src++;
			control |= (*p_src++) << 8;
		}
		unroll = p_src <= p_src_max16 ? 16 : 1;

		while (unroll--) {
			if (control & 1) {
				register UBYTE *p;
				register UCARD lenmt;
				register UBYTE *p_ziv = p_dst;
				register UCARD the_index;

				lenmt = *p_src++;
				the_index = ((lenmt & 0xF0) << 4) | *p_src++;
				p = hash[the_index];
				lenmt &= 0xF;

				*p_dst++ = *p++;
				*p_dst++ = *p++;
				*p_dst++ = *p++;
				while (lenmt--) *p_dst++ = *p++;

				if (literals > 0) {
					register UBYTE *r = p_ziv-literals;
					UPDATE_I(HASH(r), r);
					if (literals == 2) {
						r++;
						UPDATE_I(HASH(r), r);
					}
					literals = 0;
				}
				UPDATE_I(the_index & (~DEPTH_MASK), p_ziv);
			} else {
				*p_dst++ = *p_src++;
				if (++literals == 3) {
					register UBYTE *p = p_dst - 3;
					UPDATE_I(HASH(p), p);
					literals = 2;
				}
			}
			control >>= 1;
		}
	}
	*p_dst_len = p_dst - p_dst_first;
}

/* Check working memory required for each type of algorigthm */

int check_wrkmem(cmptype)
int cmptype;
{
	switch (cmptype) {
	case LZ1_COMPRESS :
		NL_LOG(80)("Working mem for lz1 algo is %lu\n", lz1_memory);
		if (lz1_memory > wrklen) {
			wrklen = lz1_memory;
			goto realloc_wrkmem;
		}
		break;
	case LZ3_3_COMPRESS :
		NL_LOG(80)("Working mem for lz3 algo is %lu\n", lz3_3_memory);
		if (lz3_3_memory > wrklen) {
			wrklen = lz3_3_memory;
			goto realloc_wrkmem;
		}
		break;
	}
	return (OUI);
realloc_wrkmem:
	if (pwrkmem) free(pwrkmem);
	if (!(pwrkmem = malloc(wrklen))) {
		NL_ERR(0)("Can't alloc work buff of %lu bytes for compression\n",
			  wrklen);
		GOTO_ERROR;
	}
	return (OUI);
error:
	return (NON);
}


void uncompress(pf)
struct st_flow *pf;
{
	UINT4 maxlen;
	ULONG compsize, realsize, len_uncompress;
	int mask;
	static UINT4 newbuf_size = 0;
	static char *newbuf = 0;
	void_funcp decompress_fn;

	mask	 = hextoi(pf->prop + 2);
	compsize = atoi(pf->value) - 5;
	realsize = atoi(pf->realsize);

	switch(mask & COMPRESS_MASK) {
	case LZ1_COMPRESS:
		if (!check_wrkmem(LZ1_COMPRESS)) GOTO_ERROR;
		decompress_fn = lz1_decompress;
		break;
	case LZ3_3_COMPRESS:
		if (!check_wrkmem(LZ3_3_COMPRESS)) GOTO_ERROR;
		decompress_fn = lz3_3_decompress;
		break;
	case NO_COMPRESS :
	default :
		NL_LOG(50)("Not compressed...\n");
		return;	/* do nothing */
	}
	maxlen = MAX(2 * compsize, realsize + 1024);
	NL_LOG(80)("Assuming memory required of %u\n", maxlen);
	if (newbuf_size < maxlen) {
		if (newbuf) free(newbuf);
		newbuf_size = maxlen;
		if (!(newbuf = malloc((unsigned int)newbuf_size))) {
			NL_ERR(0)("Can't alloc buffer for compression\n");
			newbuf_size = 0;
			GOTO_ERROR;
		}
	}
	decompress_fn(pwrkmem, pf->buff, compsize, newbuf, &len_uncompress);
	NL_LOG(80)("Uncompressed %ld bytes to %ld\n", compsize, len_uncompress);
	if (len_uncompress != realsize) {
		NL_ERR(0)("WARNING: %ld bytes expand to %ld bytes instead of %ld\n",
			  compsize, len_uncompress, realsize);
		GOTO_ERROR;
	}
	free(pf->buff);
	pf->buff = newbuf;
	pf->buff_size = len_uncompress;
	pf->flags |= DONT_FREE_BUFF;
	return ;
error:
	return ;	/* Buffer won't be uncompressed */
}
